#include "Camera.h"
#include <numbers>

void raytry::Camera::for_each_ray(const QImage &image, const std::function<void(int, int, raytry::Ray)> &consumer) {
    const auto widthf = static_cast<float>(image.width());
    const auto heightf = static_cast<float>(image.height());
    const float widthToHeight = widthf / heightf;
    const float tan = std::tan(std::max(1.0f, std::min(hfov, 179.0f)) * std::numbers::pi_v<float> / 360);
    const float xblub = 2 * planeDistance * tan / widthf;
    QVector3D zeroOrigin{
            positionVec.x() - (xblub * (widthf / 2)),
            positionVec.y(),
            positionVec.z() + (xblub * widthToHeight * (heightf / 2))};
    zeroOrigin += forwardVec * planeDistance;
    for (int y = 0; y < image.height(); ++y) {
        for (int x = 0; x < image.width(); ++x) {
            const QVector3D &directionVec = zeroOrigin + QVector3D(static_cast<float>(x) * xblub, 0.0f, -static_cast<float>(y) * xblub * widthToHeight);
            consumer(x, y, Ray(positionVec, (directionVec - positionVec).normalized()));
        }
    }
}

void raytry::Camera::for_each_bundle(const QImage &image, const std::function<void(int, int, const std::array<raytry::Ray, raytry::bundle_size>&)> &consumer) {
    const auto widthf = static_cast<float>(image.width());
    const auto heightf = static_cast<float>(image.height());
    const float widthToHeight = widthf / heightf;
    const float tan = std::tan(std::max(1.0f, std::min(hfov, 179.0f)) * std::numbers::pi_v<float> / 360);
    const float xblub = 2 * planeDistance * tan / widthf;
    const float x2blub = xblub / pixel_splits;
    QVector3D zeroOrigin{
            positionVec.x() - (xblub * (widthf / 2)),
            positionVec.y(),
            positionVec.z() + (xblub * widthToHeight * (heightf / 2))};
    zeroOrigin += forwardVec * planeDistance;
    std::array<Ray, bundle_size> bundledRays{};
    for (int y = 0; y < image.height(); ++y) {
        for (int x = 0; x < image.width(); ++x) {
            for (int y2 = 0; y2 < pixel_splits; ++y2) {
                for (int x2 = 0; x2 < pixel_splits; ++x2) {
                    float xpos = (static_cast<float>(x) * xblub) + (static_cast<float>(x2) * x2blub);
                    float zpos = ((-static_cast<float>(y) * xblub) - (static_cast<float>(y2) * x2blub)) * widthToHeight;
                    const QVector3D &directionVec = zeroOrigin + QVector3D(xpos, 0.0f, zpos);
                    bundledRays[x2 + (y2 * pixel_splits)] = Ray(positionVec, (directionVec - positionVec).normalized());
                }
            }
            consumer(x, y, bundledRays);
        }
    }
}

const QVector3D &raytry::Camera::pos() const {
    return positionVec;
}

const QVector3D &raytry::Camera::fwd() const {
    return forwardVec;
}
