#ifndef RAYTRYCPP_CAMERA_H
#define RAYTRYCPP_CAMERA_H

#include "Ray.h"
#include <QApplication>
#include <QImage>
#include <QVector3D>

namespace raytry {
    const constexpr size_t pixel_splits{2};
    const constexpr size_t bundle_size{pixel_splits * pixel_splits};

    class Camera {
    private:
        QVector3D forwardVec{0.0, 1.0, 0.0};
        QVector3D upVec{0.0, 0.0, 1.0};
        QVector3D rightVec{1.0, 0.0, 0.0};
        QVector3D positionVec{0.0, 0.0, 1.8};
        float hfov{110.0f};
        float planeDistance{0.2f};

    public:
        void for_each_ray(const QImage &image, const std::function<void(int, int, raytry::Ray)>& consumer);
        void for_each_bundle(const QImage &image, const std::function<void(int, int, const std::array<raytry::Ray, raytry::bundle_size>&)>& consumer);
        [[nodiscard]] const QVector3D& pos() const;
        [[nodiscard]] const QVector3D& fwd() const;
    };
}// namespace raytry

#endif//RAYTRYCPP_CAMERA_H
