#include "RenderMaterial.h"

#define DIVIDE_SPECULAR_BY_NDOTL

namespace raytry {
    RenderMaterial::RenderMaterial(QColor materialColor) : diffuseColor{materialColor}, specularFalloff{32.0f} {
        specularColor = addColors({204, 204, 204}, mixColor(materialColor, 0.2f));
    }

    RenderMaterial::RenderMaterial(const QColor &diffuseColor, const QColor &specularColor, float specularFalloff) : diffuseColor(diffuseColor), specularColor(specularColor),  specularFalloff(specularFalloff) {}

    QColor RenderMaterial::brdf(const QVector3D &toLight, const QVector3D &normal, const QVector3D &incomingDir, const QVector3D &outgoingDir) const {
        // source: https://github.com/wdas/brdf/blob/main/src/brdfs/blinnphong.brdf
        const QVector3D &halfway = (toLight + outgoingDir).normalized();
        float specularIntensity = std::pow(std::max(0.0f, QVector3D::dotProduct(normal, halfway)), specularFalloff);
#ifdef DIVIDE_SPECULAR_BY_NDOTL
        specularIntensity /= std::max(0.0001f, QVector3D::dotProduct(normal, toLight));
#endif
        specularIntensity = std::min(1.0f, specularIntensity);
        float diffuseIntensity = 1.0f - specularIntensity;
        return addColors(mixColor(diffuseColor, diffuseIntensity), mixColor(specularColor, specularIntensity));
    }

    QColor mixColor(QColor input, float intensity) {
        assert(intensity >= 0);
        input.setRedF(std::min(1.0f, input.redF() * intensity));
        input.setGreenF(std::min(1.0f, input.greenF() * intensity));
        input.setBlueF(std::min(1.0f, input.blueF() * intensity));
        return input;
    }

    QColor mixColors(QColor color1, const QColor& color2) {
        color1.setRedF(color1.redF() * color2.redF());
        color1.setGreenF(color1.greenF() * color2.greenF());
        color1.setBlueF(color1.blueF() * color2.blueF());
        return color1;
    }

    QColor addColors(QColor color1, const QColor& color2) {
        color1.setRedF(std::min(1.0f, color1.redF() + color2.redF()));
        color1.setGreenF(std::min(1.0f, color1.greenF() + color2.greenF()));
        color1.setBlueF(std::min(1.0f, color1.blueF() + color2.blueF()));
        return color1;
    }
}
