#ifndef RAYTRYCPP_SCENE_H
#define RAYTRYCPP_SCENE_H

#include "Camera.h"
#include "viewerwindow.h"
#include "InfinitePlane.h"
#include "KDNode.h"
#include "KDTree.h"
#include "Sphere.h"
#include "Triangle.h"
#include <QImage>

namespace raytry {
    enum IntersectionMode { full, foreground, background };

    class Scene {
    public:
        Scene();
        Scene(const Scene&) = delete;
        ~Scene() = default;
        Scene &operator=(const Scene&) = delete;

        void render(ViewerWindow& window, std::function<void()> updateImage);

    private:
        std::list<std::pair<std::reference_wrapper<const RenderObject>, float>> findIntersections(const raytry::Ray &ray, IntersectionMode mode);
        QColor renderingEquation(const RenderObject &object, const QVector3D &hitpoint, const QVector3D &outgoingDir, unsigned int depth);
        QColor incomingLight(const QVector3D &point, const QVector3D &incomingDir, unsigned int depth);
        QQuaternion nextRandomRotation();

        static const constexpr unsigned int maxDepth{4};
        static const constexpr unsigned int pathRandomDirectionCount{16};
        static const constexpr QColor backgroundColor{177, 190, 223};
        static const constexpr QVector3D pointLight{-10, -10, 10};
        static const constexpr QColor lightColor{244, 233, 155};
        size_t triangleIntersections{0};

        Camera camera{};
        InfinitePlane ground{};
        std::array<Sphere, 11> spheres{};
        std::array<Triangle, 3> triangles{};
        std::vector<Triangle> teapotMesh{};
        std::vector<std::reference_wrapper<RenderObject>> background{};
        KD::Tree foregroundTree{};
        std::random_device randomDevice{};
        std::mt19937 randomGenerator{randomDevice()};
        std::uniform_real_distribution<float> pipiDistribution{0.0f, 2 * std::numbers::pi_v<float>};
        std::uniform_real_distribution<float> halfpiDistribution{0.0f, std::numbers::pi_v<float> / 2};
    };
}// namespace raytry


#endif//RAYTRYCPP_SCENE_H
